package com.tomtom.business.android.btdemo;

import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Method;
import java.util.UUID;

import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;
import android.util.Log;

/**
 * Helper class that encapsulates our bluetooth "protocol".
 */
public class Connection
{
    private final static String TAG = "btemo.conn";
    
    // The RFCOMM channel identifier UUID of the LINK 510
    private final static String UUID_LINK510  = "a1c6e480-c101-11e1-afa7-0800200c9a66";
    
    // Fallback UUID. This is the "well known" SPP UUID supported by a wide range of BT devices
    private final static String UUID_FALLBACK = "00001101-0000-1000-8000-00805F9B34FB";
    
    // LINK 300/310 does not support channel UUIDs. We have to use the port number instead.
    private final static int RFCOMM_CHANNEL = 5;
    
    private final static String TOKEN     = "btdemo:"; // our protocol identifier

    private BluetoothDevice device = null;
    private BluetoothSocket socket = null;
    private InputStream is         = null;
    private OutputStream os        = null;
    
    /**
     * constructor to create a new connection to the given device.
     * @param device the device to connect to.
     */
    Connection(BluetoothDevice device)
    {
        this.device = device;
    }

    /**
     * Sends the given string to the LINK device.
     * @param payload the payload data.
     */
    void send(String payload)
    {
        this.init();
        
        try
        {
            byte[] data = this.encode(payload);
            Log.d(TAG,"sending data: " + new String(data));
            this.os.write(data);

            Log.d(TAG,"flushing outputstream");
            this.os.flush();
        }
        catch (IOException e)
        {
            Log.e(TAG,"error while sending data to LINK",e);
            
            // close/null on error to trigger automatic reconnect
            this.internalClose(this.os);
            this.os = null;
            
            // rethrow exception
            throw new RuntimeException(e.getMessage(),e);
        }
    }
    
    /**
     * Receives a single data block from the LINK.
     * Note: This method blocks until some data has been received.
     * @return the received data or NULL if not usable data could be received.
     */
    public String receive()
    {
        this.init();

        try
        {
            Log.d(TAG,"waiting for data from inputstream");
            
            byte[] buf = new byte[1024];
            int read = 0;
            read = this.is.read(buf);
            
            if (read > 0)
            {
                String data = new String(buf,0,read);
                Log.d(TAG,"received data: " + data);
                
                if (data.startsWith(TOKEN))
                    return data.substring(TOKEN.length());
                
                Log.w(TAG,"received unsupported data: " + data);
                return null;
            }
        }
        catch (IOException e)
        {
            Log.i(TAG,"connection closed by peer");
            Log.d(TAG,"stacktrace for debugging purpose: ",e);

            // close/null on error to trigger automatic reconnect
            this.internalClose(this.is);
            this.is = null;
        }
        
        Log.w(TAG,"received no data");
        return null;
    }
    
    /**
     * Opens the bluetooth socket and streams if needed.
     */
    private void init()
    {
        try
        {
            ///////////////////////////////////////////////
            // open socket if needed
            if (this.socket == null)
            {
                // we have to re-create both streams in this case.
                this.close();

                // 1) try to connect using LINK 510's UUID
                this.socket = this.createSocket(UUID_LINK510);
                
                // 2) fallback 1: use well-known UUID
                if (this.socket == null)
                    this.socket = this.createSocket(UUID_FALLBACK);
                
                // 3) fallback 2: LINK 300/310 - use port number
                if (this.socket == null)
                    this.socket = this.createSocket(RFCOMM_CHANNEL);
                
                if (this.socket == null)
                    throw new IOException("unable to connect to bluetooth device");
                
            }
            //
            ///////////////////////////////////////////////
            
            ///////////////////////////////////////////////
            // open input stream if needed
            if (this.is == null)
            {
                Log.d(TAG,"opening inputstream");
                this.is = this.socket.getInputStream();
            }
            //
            ///////////////////////////////////////////////
            
            ///////////////////////////////////////////////
            // open output stream if needed
            if (this.os == null)
            {
                Log.d(TAG,"opening outputstream");
                this.os = this.socket.getOutputStream();
            }
            //
            ///////////////////////////////////////////////
        }
        catch (IOException e)
        {
            Log.e(TAG,"unable to open BT socket and/or streams",e);
            
            // make sure, the socket and streams are NULL to be able to do a re-connect
            this.close();
            
            // rethrow exception
            throw new RuntimeException(e.getMessage(),e);
        }
    }
    
    /**
     * Tries to create the bluetooth socket by UUID.
     * This method tries to connect to the socket before returning it - just to make sure, it works.
     * @param uuid the UUID.
     * @return the socket or NULL in case of an error.
     */
    private BluetoothSocket createSocket(String uuid)
    {
        try
        {
            UUID id = UUID.fromString(uuid);
            Log.i(TAG,"try to open BT socket using service UUID " + id);
            BluetoothSocket socket = this.device.createInsecureRfcommSocketToServiceRecord(id);
            Log.i(TAG,"connecting to socket");
            socket.connect();
            return socket;
        }
        catch (Exception e)
        {
            Log.d(TAG,"could not open socket",e);
        }
        
        return null;
    }
    
    /**
     * Tries to create the bluetooth socket by port number.
     * This method tries to connect to the socket before returning it - just to make sure, it works.
     * @param port the port number.
     * @return the socket or NULL in case of an error.
     */
    private BluetoothSocket createSocket(int port)
    {
        try
        {
            // unfortunately this method is not part of the public Android API.
            // therefore we have to use a reflection call.
            Log.i(TAG,"try to open BT socket using port " + port);
            Method m = this.device.getClass().getMethod("createInsecureRfcommSocket", new Class[] {int.class});
            BluetoothSocket socket = (BluetoothSocket) m.invoke(this.device, port);
            Log.i(TAG,"connecting to socket");
            socket.connect();
            return socket;
        }
        catch (Exception e)
        {
            Log.d(TAG,"could not open socket",e);
        }
        
        return null;
    }
    
    /**
     * Encodes the given payload data.
     * @param command the command to send.
     * @param payload the payload.
     * @return the bytes to send.
     * @throws UnsupportedEncodingException
     */
    private byte[] encode(String payload) throws UnsupportedEncodingException
    {
        String data = TOKEN + payload; // prepend our protocol token
        return data.getBytes();
    }
    
    /**
     * Closes the bluetooth socket and the streams.
     */
    public void close()
    {
        try
        {
            this.internalClose(this.is,this.os,this.socket);
        }
        finally
        {
            this.is     = null;
            this.os     = null;
            this.socket = null;
        }
    }
    
    /**
     * Closes the given resource(s) silently.
     * @param list one or more resources to close.
     */
    private void internalClose(Closeable... list)
    {
        if (list == null || list.length == 0)
            return;
        
        for (Closeable c:list)
        {
            if (c == null)
                continue;
            
            String name = c.getClass().getSimpleName();
            try
            {
                Log.d(TAG,"closing resource " + name);
                c.close();
            }
            catch (Exception e)
            {
                Log.e(TAG,"error while closing " + name,e);
            }
        }
    }
}
