package com.tomtom.business.android.btdemo;

import java.util.ArrayList;

import android.app.Activity;
import android.app.ProgressDialog;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.IntentFilter;
import android.os.Bundle;
import android.util.Log;

/**
 * The LINK discovery activity.
 */
public class Discovery extends Activity
{
    public final static String INTENT_RESULT = Discovery.class.getName() + ".RESULT";
    
    private final static String TAG = "btemo.discovery";
    
    private ProgressDialog wait        = null;
    private BroadcastReceiver receiver = null;
    
    private ArrayList<BluetoothDevice> result = new ArrayList<BluetoothDevice>();
    private int status = Activity.RESULT_OK;
    
    /**
     * @see android.app.Activity#onCreate(android.os.Bundle)
     */
    @Override
    protected void onCreate(Bundle state)
    {
        super.onCreate(state);
        
        Log.i(TAG,"start BT discovery");
        
        // register for BT discovery events
        this.receiver = new MyReceiver();
        this.registerReceiver(this.receiver,new IntentFilter(BluetoothAdapter.ACTION_DISCOVERY_FINISHED));
        this.registerReceiver(this.receiver,new IntentFilter(BluetoothDevice.ACTION_FOUND));

        
        // create wait dialog
        this.wait = ProgressDialog.show(this,this.getText(R.string.status_discovery),this.getText(R.string.wait));
        this.wait.setCancelable(true);
        this.wait.setOnCancelListener(new DialogInterface.OnCancelListener()
        {
            public void onCancel(DialogInterface dialog)
            {
                Log.i(TAG,"BT discovery cancelled");
                status = Activity.RESULT_CANCELED;
                finish();
            }
        });

        // cancel possibly running discovery
        BluetoothAdapter adapter = BluetoothAdapter.getDefaultAdapter();
        if (adapter.isDiscovering())
        {
            Log.i(TAG,"cancel currently running BT discovery");
            adapter.cancelDiscovery();
        }
        
        // start new discovery
        Log.i(TAG,"start new BT discovery");
        adapter.startDiscovery();
    }

    /**
     * @see android.app.Activity#finish()
     */
    @Override
    public void finish()
    {
        // unregister broadcast receiver
        if (this.receiver != null)
            this.unregisterReceiver(this.receiver);

        // close wait dialog
        if (this.wait != null && this.wait.isShowing()) // true if not cancelled
            this.wait.dismiss();

        // pass discovery result back to main activity
        Intent intent = this.getIntent();
        intent.putParcelableArrayListExtra(INTENT_RESULT,result);
        
        this.setResult(this.status,intent);
        super.finish();
    }

    /**
     * Receives BT discovery events.
     */
    private class MyReceiver extends BroadcastReceiver
    {
        /**
         * @see android.content.BroadcastReceiver#onReceive(android.content.Context, android.content.Intent)
         */
        @Override
        public void onReceive(Context ctx, Intent intent)
        {
            String action = intent.getAction();
            
            //////////////////////////////////////////////////////
            // discovery finished
            if (BluetoothAdapter.ACTION_DISCOVERY_FINISHED.equals(action))
            {
                Log.i(TAG,"BT discovery finished");
                finish();
                return;
            }
            //
            //////////////////////////////////////////////////////
            
            //////////////////////////////////////////////////////
            // BT device found
            if (BluetoothDevice.ACTION_FOUND.equals(action))
            {
                BluetoothDevice device = intent.getParcelableExtra(BluetoothDevice.EXTRA_DEVICE);
                Log.i(TAG,"BT discovery: found " + device.getAddress());
                if (this.isLINK(device))
                {
                    Log.i(TAG,"is LINK device, adding to result list");
                    result.add(device);
                    wait.setMessage(getString(R.string.status_discovery_count,result.size()));
                }
                return;
            }
            //
            //////////////////////////////////////////////////////
        }
        
        /**
         * Checks if the given device is a LINK.
         * @param device the device to check.
         * @return true if the device is a LINK, otherwise false.
         */
        private boolean isLINK(BluetoothDevice device)
        {
            // 1) check BT address. Has to start with "00:13:6C" or "00:21:3E"
            String address = device.getAddress();
            if (!address.startsWith("00:13:6C") && !address.startsWith("00:21:3E"))
                return false;
            
            // 2) check if device name starts with "LINK"
            String name = device.getName();
            if (name == null)
                name = device.getName(); // 2nd attempt - seems to be a race condition somewhere
            return name != null && name.startsWith("LINK");
        }
    }

}
