package com.tomtom.business.android.btdemo;

import java.util.ArrayList;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.ProgressDialog;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.res.Resources;
import android.os.AsyncTask;
import android.os.Bundle;
import android.telephony.TelephonyManager;
import android.util.Log;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.TextView;
import android.widget.Toast;


/**
 * Main activity.
 * This is the primary window of the app.
 */
public class Main extends Activity
{
    private final static String TAG = "btemo.main";
    
    private static final int REQ_BT_ENABLE      = 1;
    private static final int REQ_LINK_DISCOVERY = 2;
    private static final int REQ_LINK_CHOOSE    = 3;

    private Settings settings        = null; // our settings
    private BluetoothAdapter adapter = null; // the local bluetooth adapter
    private BluetoothDevice link     = null; // connection to LINK device
    
    private ResponseBroadcastReceiver receiver = new ResponseBroadcastReceiver(); // receives broadcast events from the response thread
    private ResponseThread response            = null; // thread that waits for server responses

    /**
     * @see android.app.Activity#onCreate(android.os.Bundle)
     */
    @Override
    public void onCreate(Bundle savedInstanceState)
    {
        Log.i(TAG,"create activity");
        super.onCreate(savedInstanceState);

        this.settings = new Settings(this.getSharedPreferences(TAG,Context.MODE_PRIVATE));
        
        this.setContentView(R.layout.main);
        ((TextView) this.findViewById(R.id.imei)).setText(this.getIMEI());
        
        this.setupBT();
    }
    
    /**
     * @see android.app.Activity#onStart()
     */
    @Override
    protected void onStart()
    {
        super.onStart();
        this.registerReceiver(this.receiver,this.receiver.getFilter());
    }
    
    /**
     * @see android.app.Activity#onStop()
     */
    @Override
    protected void onStop()
    {
        super.onStop();
        this.unregisterReceiver(this.receiver);
    }
    
    /**
     * Sets up the bluetooth adapter.
     */
    private void setupBT()
    {
        if (this.adapter != null)
            return; // already initialized
        
        Log.i(TAG,"checking BT adapter");
        this.adapter = BluetoothAdapter.getDefaultAdapter();
        if (this.adapter == null)
            this.error(R.string.err_bt_notsupported);
        
        Log.i(TAG,"checking BT state");
        if (this.adapter.isEnabled())
        {
            this.setupLINK();
        }
        else
        {
            Log.d(TAG,"BT disabled, ask user to enable");
            this.startActivityForResult(new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE),REQ_BT_ENABLE);
        }
    }

    /**
     * Sets up the BT connection to the LINK device.
     * If there's no pairing, this method automatically initiates a device discovery.
     */
    private void setupLINK()
    {
        if (this.link != null)
            return; // already initialized

        // display the phone's BT address
        String localAddress = this.adapter.getAddress();
        String localName    = this.adapter.getName();
        Log.i(TAG,"local BT address: " + localAddress);
        Log.i(TAG,"local BT friendly name: " + localName);
        ((TextView) this.findViewById(R.id.bt_local)).setText(localAddress);
        ((TextView) this.findViewById(R.id.bt_name)).setText(localName);
        
        Log.i(TAG,"setup LINK device");
        String address = this.settings.get(Settings.LINK_ADDRESS,null);
        if (address != null)
        {
            Log.i(TAG,"found stored BT address: " + address + ", choosing this");
            this.choose(this.adapter.getRemoteDevice(address));
        }
        else
        {
            // no LINK paired. start device discovery
            if (this.adapter.isDiscovering())
            {
                Log.i(TAG,"cancel currently running BT discovery");
                this.adapter.cancelDiscovery();
            }
            this.startActivityForResult(new Intent(this,Discovery.class),REQ_LINK_DISCOVERY);
        }
        
    }
    
    /**
     * Chooses the given LINK device.
     * @param device the device to choose.
     */
    private void choose(BluetoothDevice device)
    {
        this.link = device;
        String address = device.getAddress();
        String name    = device.getName();

        // store in settings
        this.settings.set(Settings.LINK_ADDRESS,address);

        // update status text
        this.status(R.string.status_chosen,name != null ? name : "LINK",address);
        
        // enable assign button
        this.setButtonEnabled(true);
    }
    
    /**
     * Performs the driver assignment.
     * @param v the according button, that requested the action (see "android:onClick" in layout/main.xml).
     */
    public void assign(View v)
    {
        // just to make sure, we have a link
        if (this.link == null)
        {
            Toast.makeText(this,R.string.status_not_chosen,Toast.LENGTH_SHORT).show();
            return;
        }
        
        // disable the assign button to prevent subsequent clicks
        this.setButtonEnabled(false);
        
        // stop a possibly running ResponseThread. The thread internally closes the connection too
        if (this.response != null && !this.response.isInterrupted())
            this.response.interrupt();
        
        // open new connection
        final Connection conn = new Connection(link);
        
        // Execute the bluetooth communication as AsyncTask
        AsyncTask<Void,Void,Exception> task = new AsyncTask<Void,Void,Exception>()
        {
            private ProgressDialog dialog = null;
            
            /**
             * @see android.os.AsyncTask#onPreExecute()
             */
            @Override
            protected void onPreExecute()
            {
                super.onPreExecute();
                Context ctx = Main.this;
                this.dialog = ProgressDialog.show(ctx,ctx.getText(R.string.status_assigning),
                                                      ctx.getText(R.string.wait));
            }

            /**
             * @see android.os.AsyncTask#doInBackground(Params[])
             */
            @Override
            protected Exception doInBackground(Void... params)
            {
                try
                {
                    Log.i(TAG,"BEGIN");
                    conn.send(getIMEI());
                    Log.i(TAG,"END");
                    return null;
                }
                catch (Exception e)
                {
                    return e;
                }
            }
            
            /**
             * @see android.os.AsyncTask#onPostExecute(java.lang.Object)
             */
            @Override
            protected void onPostExecute(Exception result)
            {
                super.onPostExecute(result);
                this.dialog.dismiss();
                
                if (result != null) // bluetooth communication failed
                {
                    status(R.string.status_assign_error,result.getMessage());
                    // re-enable assign button
                    setButtonEnabled(true);
                }
                else // bluetooth communication successful. Start background thread that handles the server response
                {
                    status(R.string.status_assign_success);
                    
                    // Start new backround thread that waits for the servers response
                    response = new ResponseThread(Main.this,conn);
                    response.start();
                }
            }
        };
        task.execute();
    }
    
    /**
     * @see android.app.Activity#onActivityResult(int, int, android.content.Intent)
     */
    @Override
    protected void onActivityResult(int request, int result, Intent intent)
    {
        Log.i(TAG,"got activity result " + result + " for request " + request);
        super.onActivityResult(request, result, intent);

        /////////////////////////////////////
        // coming back from enabling BT
        if (request == REQ_BT_ENABLE)
        {
            if (result != Activity.RESULT_OK)
            {
                Log.e(TAG,"bluetooth not enabled");
                this.error(R.string.err_bt_notenabled);
            }
            
            // we can now setup the LINK
            this.setupLINK();
            return;
        }
        //
        /////////////////////////////////////
        
        /////////////////////////////////////
        // coming back from LINK discovery
        if (request == REQ_LINK_DISCOVERY)
        {
            if (result == Activity.RESULT_CANCELED)
            {
                Log.i(TAG,"BT LINK discovery cancelled by user");
                
                // if we're already connected to a LINK, we leave the status unchanged,
                if (this.link == null)
                    this.status(R.string.status_discovery_cancelled);
                
                return;
            }

            ArrayList<BluetoothDevice> devices = intent.getParcelableArrayListExtra(Discovery.INTENT_RESULT);
            
            // Did not found any LINK devices
            if (devices == null || devices.size() == 0)
            {
                Log.w(TAG,"no LINK devices found");
                
                // if we're already connected to a LINK, we don't overwrite the status, instead we just show a toast.
                if (this.link != null)
                    Toast.makeText(this,R.string.status_discovery_notfound,Toast.LENGTH_SHORT).show();
                else
                    this.status(R.string.status_discovery_notfound);
                
                return;
            }
            
            // found exactly one. Choose this one automatically - but only if we're not yet connected to a LINK
            if (devices.size() == 1 && this.link == null)
            {
                choose(devices.get(0));
                return;
            }
            
            // ask the user which LINK to use
            Intent i = new Intent(this,Choice.class);
            i.putParcelableArrayListExtra(Choice.INTENT_LIST,devices);
            i.putExtra(Choice.INTENT_CURRENT,this.link); // used to preselect the current LINK
            this.startActivityForResult(i,REQ_LINK_CHOOSE);
        }
        //
        /////////////////////////////////////

        /////////////////////////////////////
        // coming back from LINK choice
        if (request == REQ_LINK_CHOOSE)
        {
            if (result == Activity.RESULT_CANCELED)
            {
                Log.i(TAG,"LINK choice cancelled by user");
                return;
            }
            
            BluetoothDevice device = intent.getParcelableExtra(Choice.INTENT_RESULT);
            if (device == null)
            {
                Log.i(TAG,"no LINK device choosen by user");
                return;
            }
            
            this.choose(device);
        }
        //
        /////////////////////////////////////
    }

    /**
     * Returns the phone's IMEI to be able to identify the device on the server side.            
     * @return the phone's IMEI.
     */
    private String getIMEI()
    {
        TelephonyManager tm = (TelephonyManager) this.getSystemService(Context.TELEPHONY_SERVICE);
        return tm.getDeviceId();
    }
    
    /**
     * Enables/disables the "assign" button.
     * @param enabled true to enable, fals to disable.
     */
    private void setButtonEnabled(boolean enabled)
    {
        this.findViewById(R.id.assign).setEnabled(enabled);
    }

    /**
     * Updates the status text in the UI.
     * @param status the status text (from R.string).
     */
    private void status(int status)
    {
        this.status(status,(Object[])null);
    }
    
    /**
     * Updates the status text in the UI.
     * @param status the status text (from R.string).
     * @param params list of additional format args.
     */
    private void status(int status, Object... args)
    {
        ((TextView) this.findViewById(R.id.status)).setText(this.getString(status,args));
    }
    
    /**
     * Shows an alert with the given reason and exits the app.
     * @param reason the reason (from R.string).
     */
    private void error(int reason)
    {
        Resources res = this.getResources();
        String msg = res.getString(reason);
        Log.w(TAG,"exiting, reason: " + msg);
        
        AlertDialog d = new AlertDialog.Builder(this).create();
        d.setTitle(res.getString(R.string.sorry));
        d.setMessage(msg);
        d.setButton(DialogInterface.BUTTON_NEGATIVE,res.getString(R.string.close),new DialogInterface.OnClickListener()
        {
            public void onClick(DialogInterface dialog, int which)
            {
                finish();
            }
        });
        d.show();
    }
    
    /**
     * @see android.app.Activity#onCreateOptionsMenu(android.view.Menu)
     */
    @Override
    public boolean onCreateOptionsMenu(Menu menu)
    {
        super.onCreateOptionsMenu(menu);
        this.getMenuInflater().inflate(R.menu.main, menu);
        return true;
    }

    /**
     * @see android.app.Activity#onMenuItemSelected(int, android.view.MenuItem)
     */
    @Override
    public boolean onMenuItemSelected(int featureId, MenuItem item)
    {
        // start activity to re-discover LINK devices
        if (item.getItemId() == R.id.menu_scan)
        {
            this.startActivityForResult(new Intent(this,Discovery.class),REQ_LINK_DISCOVERY);
            return true;
        }
        return super.onMenuItemSelected(featureId, item);
    }
    
    /**
     * Receives broadcast events from the response thread.
     */
    private class ResponseBroadcastReceiver extends BroadcastReceiver
    {
        private IntentFilter filter = null;
        
        /**
         * @see android.content.BroadcastReceiver#onReceive(android.content.Context, android.content.Intent)
         */
        @Override
        public void onReceive(Context context, Intent intent)
        {
            // re-enable button
            setButtonEnabled(true);

            String action = intent.getAction();
            
            if (action != null && action.equals(Settings.INTENT_RESPONSE_TIMEOUT))
                status(R.string.status_assign_response_timeout);
            else if (action != null && action.equals(Settings.INTENT_RESPONSE_OK))
                status(R.string.status_assign_response_ok,intent.getCharSequenceExtra(Settings.INTENT_RESPONSE_OK));
            else
                Log.w(TAG,"received unknown intent " + intent);
        }
        
        /**
         * Returns the intent filter we are listening to.
         * @return the intent filter we are listening to.
         */
        private IntentFilter getFilter()
        {
            if (this.filter == null)
            {
                this.filter = new IntentFilter();
                this.filter.addAction(Settings.INTENT_RESPONSE_OK);
                this.filter.addAction(Settings.INTENT_RESPONSE_TIMEOUT);
            }
            return filter;
        }
    }
}
