package com.tomtom.business.android.btdemo;

import android.content.Context;
import android.content.Intent;
import android.util.Log;


/**
 * Background thread, that waits for the servers response.
 */
public class ResponseThread extends Thread
{
    private final static String TAG = "btemo.response";
    
    /**
     * The number of seconds we will wait for the servers response.
     */
    private final static int TIMEOUT_SECONDS = 30;
    
    private Context ctx     = null;
    private Connection conn = null;
    
    /**
     * constructor.
     * @param ctx the context.
     * @param conn the bluetooth connection to use.
     */
    public ResponseThread(Context ctx, Connection conn)
    {
        super("[" + TAG + "] response thread");
        this.ctx  = ctx;
        this.conn = conn;
    }
    
    /**
     * @see java.lang.Thread#run()
     */
    @Override
    public void run()
    {
        // We'll wait for about a minute for the servers response
        WatcherThread watcher = new WatcherThread();
        watcher.start();

        // start waiting for response
        try
        {
            while (true)
            {
                Log.i(TAG,"waiting for server response ...");
                String data = this.conn.receive();
                if (data != null)
                {
                    Log.i(TAG,"received server response: " + data);
                    
                    // notify UI using broadcast event
                    Log.i(TAG,"notifying UI using broadcast");
                    Intent broadcast = new Intent(Settings.INTENT_RESPONSE_OK);
                    broadcast.putExtra(Settings.INTENT_RESPONSE_OK,data);
                    this.ctx.sendBroadcast(broadcast);
                    
                    // watcher no longer needed. Can be interrupted too
                    watcher.interrupt();
                    return;
                }
                
                sleep(5000L); // wait a bit
            }
        }
        catch (InterruptedException e)
        {
            Log.i(TAG,"got no server response, giving up");
        }
        finally
        {
            // make sure to close the connection
            this.conn.close();
        }
    }
    
    /**
     * A watcher thread, that stops the response thread after a given timeout.
     */
    private class WatcherThread extends Thread
    {
        /**
         * constructor.
         */
        private WatcherThread()
        {
            super("[" + TAG + "] watcher thread");
        }
        
        /**
         * @see java.lang.Thread#run()
         */
        @Override
        public void run()
        {
            try
            {
                sleep(TIMEOUT_SECONDS * 1000L);
            }
            catch (InterruptedException e)
            {
                Log.i(TAG,"watcher thread interrupted");
                return;
            }

            ResponseThread r = ResponseThread.this;
            if (r.isAlive())
            {
                Log.i(TAG,"got no server response within " + TIMEOUT_SECONDS + " seconds, stopping response thread");
                
                // send interrupt to response thread 
                r.interrupt();
                
                // addintionally close the connection in case the response thread
                // hangs uninterruptible while reading from the inputstream
                conn.close();
                
                // notify UI using broadcast event
                Log.i(TAG,"notifying UI using broadcast");
                Intent broadcast = new Intent(Settings.INTENT_RESPONSE_TIMEOUT);
                ctx.sendBroadcast(broadcast);
            }
        }
    }
}
