<?php


include("config.php");
include("webfleet/include.php");

// overwrite config settings with local values if exists
if (file_exists("config.local.php")) {
    include("config.local.php");
    Server::log("Reading deployment specific configuration from config.local.php ...");
}

// perform SSl checl on local PHP install
$w = stream_get_wrappers();
Server::log("WEBFLEET.connect access through SSL: ".$config["is_ssl"]);
Server::log("OpenSSL Extension loaded: ".(extension_loaded  ('openssl') ? 'yes':'no'));
Server::log("HTTP wrapper supported: ".(in_array('http', $w) ? 'yes':'no'));
Server::log("HTTPS wrapper supported: ".(in_array('https', $w) ? 'yes':'no'));
Server::log("List of available wrappers: ".print_r($w, true), false);

// our custom protocol identifier. Do not change unless you change it in the Android app too
define("BT_TOKEN","btdemo:");

try
{
    $server = new Server();
    $server->run();
}
catch (Exception $e)
{
    print_r($e);
}

/**
 * Implements the server part of the bluetooth demo.
 */
class Server
{
    private $conn = null;

    /**
     * constructor.
     */
    public function __construct()
    {
        global $config;

        $auth    = new \webfleet\AuthParams($config["account"],$config["username"],$config["password"]);
        if(strlen(trim($config["apikey"]))>0) {
          // if we have an API Key
          $auth->apikey = $config["apikey"];
          $this->log("Using API-Key: ".$auth->apikey);
        }
        //driver mapping
        $dc = count($config["drivers"]);
        if($dc>0) {
        $this->log("Number of IMEI to driver mappings: ".$dc);
        $this->log("Dump of driver mapping: ".print_r($config["drivers"], true), false);
        } else {
        	$this->log("No IMEI to driver mapping defined!");
        }

        $general = new \webfleet\GeneralParams($config["lang"]);

        $this->log("connecting to ".$config["url"]);
        $this->conn = webfleet::connect(WF_PROTO_CSV,$config["url"],$auth,$general);
        $this->conn->proxyHostname = $config["proxy_hostname"];
        $this->conn->proxyPort     = $config["proxy_port"];
        $this->conn->proxyUsername = $config["proxy_username"];
        $this->conn->proxyPassword = $config["proxy_password"];
    }

    /**
     * The server loop.
     */
    public function run()
    {
        global $config;

        // 1) check if our queue exists
        $this->checkQueue();

        // 2) start server loop
        $this->log("start server loop, interval: ".$config["interval"]." seconds, press <CTRL><C> to stop");
        while (true)
        {
            try
            {
                // pop messages
                $result = $this->conn->execute(new \webfleet\Action("popQueueMessagesExtern",array("msgclass" => 15)));
                $this->log("fetched ".sizeof($result)." messages");

                // processing messages
                foreach ($result as $item)
                {
                    $this->process($item);
                }

                // acknowledge
                $this->conn->execute(new \webfleet\Action("ackQueueMessagesExtern",array("msgclass" => 15)));
            }
            catch (\webfleet\WebfleetException $e)
            {
                $this->log($e->getErrorCode().": ".$e->getMessage());
            }

            sleep($config["interval"]);
        }
    }

    /**
     * Creates the message queue if needed.
     */
    private function checkQueue()
    {
        try
        {
            $this->log("create message queue if needed");
            $this->conn->execute(new \webfleet\Action("createQueueExtern",array("msgclass" => 15))); // "15" is the message class for end-to-end payload messages
            $this->log("created message queue");
        }
        catch (\webfleet\WebfleetException $e)
        {
            if ($e->getErrorCode() == "WFCQ_E0006")
            {
                $this->log("queue already exists");
                return;
            }

            // rethrow if error code != WFCQ_E0006 (queue already exists)
            throw $e;
        }
    }

    /**
     * process a single message.
     * @param stdClass $msg
     */
    private function process($msg)
    {
        global $config;

        $msgType = $msg->msg_type;
        $object  = $msg->objectuid;

        // we're only interested in aux_device_data_in here
        if ($msgType != "80000862")
            return;

        if (!$object)
            return; // we dont have an object

        // extract "surplus_data" from item
        $surplus = $msg->surplus_data;
        if (!$surplus) // skip if empty
            return;

        // parse as json
        $json = json_decode($surplus);
        if (!$json)
            return;

        $data = $json->aux_device_data_in;
        if (!$data)
            return;

        // extract base64 encoded payload
        $base64 = $data->aux_data;
        if (!$base64)
            return;

        $payload = base64_decode($base64);

        // check if this is a driver assignment message
        if (strpos($payload,BT_TOKEN) !== 0)
            return;

        // extract IMEI/phone number from payload
        $phone = trim(substr($payload,strlen(BT_TOKEN)));

        // do we have a driver assigned?
        $driver = $config["drivers"][$phone];
        if (!$driver)
            return;

        // OK, we have driver + object. Try to assign.
        $this->assign($driver,$object,$data->device_id);
    }

    /**
     * Tries to assign the given driverno to the object uid.
     * @param string $driver driverno.
     * @param string $object the objectuid
     * @param string $device bluetooth address of the device, the request came from.
     */
    private function assign($driver, $object, $device)
    {
        global $config;

        $this->log("try to assign driverno ".$driver." to objectuid ".$object);

        // 1) check if the object is available
        $list = $this->conn->execute(new \webfleet\Action("showObjectReportExtern",array("objectuid" => $object)));
        $current = $list[0]->driver;
        if ($current)
        {
            // object already assigned to a driver
            if ($current == $driver)
            {
                $this->log("object already assigned");
                // notify client
                $this->sendAux($object,$device,"driver ".$driver." already assigned");
                return;
            }
            else
            {
                // may fail if the driver assigned himself using a PND or digital tachograph
                $this->log("try to unassign driver ".$current." from current object ".$object);
                $this->conn->execute(new \webfleet\Action("detachDriverFromVehicle",array("objectuid" => $object,"driverno" => $current)));
            }
        }

        // 2) check if the driver is available
        $list = $this->conn->execute(new \webfleet\Action("showDriverReportExtern",array("driverno" => $driver)));
        $current = $list[0]->objectuid;
        if ($current)
        {
            // we don't need to compare "$current == $object" - case already handled above

            $this->log("try to unassign current driver ".$driver." from object ".$current);
            $this->conn->execute(new \webfleet\Action("detachDriverFromVehicle",array("objectuid" => $current,"driverno" => $driver)));
        }

        // 3) now try to assign driver to object
        $this->log("assign driver ".$driver." to object ".$object);
        $this->conn->execute(new \webfleet\Action("attachDriverToVehicle",array("objectuid" => $object,"driverno" => $driver)));

        // 4) notify client about driver assignment
        $this->sendAux($object,$device,"successfully assigned as driver no: ".$driver);
    }

    /**
     * Sends the given payload back to the bluetooth device.
     * @param string $object the objectuid
     * @param string $device bluetooth address of the device, the request came from.
     * @param unknown_type $payload the payload to send.
     */
    private function sendAux($object, $device, $payload)
    {
        global $config;

        $this->log("sending AUX data to objectuid ".$object.", device: ".$device.", payload: ".$payload);
        $data = base64_encode(BT_TOKEN.$payload);
        $this->conn->execute(new \webfleet\Action("sendAuxDeviceData",array("objectuid" => $object,"deviceid" => $device, "data" => $data)));
    }

/**
 * @param $msg - the message to be logged
 * @param $linebreak - include line break after log message
 */
    public static function log($msg, $linebreak=true)
    {
        print("LINK.connect-Server-[".date("Y-m-d H:i:s")."] ".$msg.($linebreak?"\n":""));
    }
}

?>